"""URL userinfo normalization."""

from __future__ import annotations


def normalize_userinfo(userinfo: str) -> str:
    """Normalize userinfo part of the url.

    Params:
        userinfo : string : url userinfo, e.g., 'user@'

    Returns:
        string : normalized userinfo data.

    """
    if userinfo in ["@", ":@"]:
        return ""
    return userinfo



"""URL normalization tools."""

from __future__ import annotations

import re
import unicodedata
from typing import NamedTuple
from urllib.parse import quote as quote_orig
from urllib.parse import unquote as unquote_orig
from urllib.parse import urlsplit, urlunsplit


class URL(NamedTuple):
    """URL components tuple.

    A named tuple containing the parsed components of a URL:
    scheme, userinfo, host, port, path, query, and fragment.
    """

    scheme: str
    userinfo: str
    host: str
    port: str
    path: str
    query: str
    fragment: str


def deconstruct_url(url: str) -> URL:
    """Transform the url into URL structure.

    Params:
        url : string : the URL

    Returns:
        URL

    """
    scheme, auth, path, query, fragment = urlsplit(url.strip())
    match = re.search(r"([^@]*@)?([^:]*):?(.*)", auth)
    (userinfo, host, port) = match.groups()  # type: ignore  # noqa: PGH003
    return URL(
        fragment=fragment,
        host=host,
        path=path,
        port=port or "",
        query=query,
        scheme=scheme,
        userinfo=userinfo or "",
    )


def reconstruct_url(url: URL) -> str:
    """Reconstruct string url from URL.

    Params:
        url : URL object instance

    Returns:
        string : reconstructed url string

    """
    auth = (url.userinfo or "") + url.host
    if url.port:
        auth += ":" + url.port
    return urlunsplit((url.scheme, auth, url.path, url.query, url.fragment))


def force_unicode(string: str | bytes, charset: str = "utf-8") -> str:
    """Ensure string is properly encoded (Python 3 only).

    Params:
        string : str : an input string
        charset : str : optional : output encoding

    Returns:
        str

    """
    if isinstance(string, bytes):
        return string.decode(charset, "replace")
    return string


def unquote(string: str, charset: str = "utf-8") -> str:
    """Unquote and normalize unicode string.

    Params:
        string : string to be unquoted
        charset : string : optional : output encoding

    Returns:
        string : an unquoted and normalized string

    """
    string = unquote_orig(string)
    string = force_unicode(string, charset)
    encoded_str = unicodedata.normalize("NFC", string).encode(charset)
    return encoded_str.decode(charset)


def quote(string: str, safe: str = "/") -> str:
    """Quote string.

    Params:
        string : string to be quoted
        safe : string of safe characters

    Returns:
        string : quoted string

    """
    return quote_orig(string, safe)



"""URL query normalization."""

from __future__ import annotations

from .param_allowlist import get_allowed_params
from .tools import quote, unquote

QUERY_PARAM_SAFE_CHARS = "~:/?[]@!$'()*+,;"


def process_query_param(param: str) -> str:
    """Process a single query parameter.

    This function normalizes the parameter by quoting reserved characters
    and ensuring the parameter is in the correct format.

    Params:
        param: The query parameter to process.

    Returns:
        str: The normalized query parameter.

    """
    if not param:
        return ""
    return quote(unquote(param), QUERY_PARAM_SAFE_CHARS)


def normalize_query(
    query: str,
    *,  # Force keyword-only arguments
    host: str | None = None,
    filter_params: bool = False,
    param_allowlist: list | dict | None = None,
) -> str:
    """Normalize query while preserving parameter order.

    Params:
        query: URL query string (e.g. 'param1=val1&param2')
        host: Domain for allowlist checks
        filter_params: If True, removes non-allowlisted parameters
        param_allowlist: Optional override for default allowlist

    Returns:
        Normalized query string with original parameter order

    """
    if not query:
        return ""

    processed = []
    for param in query.split("&"):
        if not param:
            continue
        key, _, value = param.partition("=")
        key = process_query_param(key)
        if filter_params:
            allowed_params = get_allowed_params(host, param_allowlist)
            if key not in allowed_params:
                continue
        value = process_query_param(value)
        processed.append(f"{key}={value}" if value else key)

    return "&".join(processed)



"""URL scheme validation and attachment."""

from __future__ import annotations

from .normalize_scheme import DEFAULT_SCHEME

# Schemes that require authority component reconstruction with //
AUTHORITY_SCHEMES = frozenset(["http", "https", "ftp", "ftps"])


def provide_url_scheme(url: str, default_scheme: str = DEFAULT_SCHEME) -> str:
    """Make sure we have valid url scheme.

    Params:
        url : string : the URL
        default_scheme : string : default scheme to use, e.g. 'https'

    Returns:
        string : updated url with validated/attached scheme

    """
    has_scheme = ":" in url[:7]
    is_universal_scheme = url.startswith("//")
    is_file_path = url == "-" or (url.startswith("/") and not is_universal_scheme)
    if not url or is_file_path:
        return url
    if not has_scheme:
        return f"{default_scheme}://{url.lstrip('/')}"
    scheme_part, rest = url.split(":", 1)
    if scheme_part.lower() not in AUTHORITY_SCHEMES:
        # handle cases like tel:, mailto:, etc.
        return url
    return f"{scheme_part}://{rest.lstrip('/')}"



"""URL domain validation and attachment."""

from __future__ import annotations


def provide_url_domain(url: str, default_domain: str | None = None) -> str:
    """Add default domain to URL if needed.

    For absolute paths (starting with '/'), adds the specified default domain.

    Params:
        url : str : the URL
        default_domain : str | None : default domain to use, e.g. 'example.com'

    Returns:
        str : URL with domain added if applicable

    """
    # Skip processing if no default domain provided or URL is empty or stdout
    if not default_domain or not url or url == "-":
        return url

    # Only apply to absolute paths (starting with '/')
    # but not scheme-relative URLs ('//')
    if url.startswith("/") and not url.startswith("//"):
        return "//" + default_domain + url

    return url



"""URI normalize.

Copyright (c) 2020 Nikolay Panov
SPDX-License-Identifier: MIT

"""

from .url_normalize import url_normalize

__license__ = "MIT"
__version__ = "2.2.1"

__all__ = ["url_normalize"]



"""URL port normalization."""

from __future__ import annotations

DEFAULT_PORT = {
    "ftp": "21",
    "gopher": "70",
    "http": "80",
    "https": "443",
    "news": "119",
    "nntp": "119",
    "snews": "563",
    "snntp": "563",
    "telnet": "23",
    "ws": "80",
    "wss": "443",
}


def normalize_port(port: str, scheme: str) -> str:
    """Normalize port part of the url.

    Remove mention of default port number

    Params:
        port : string : url port, e.g., '8080'
        scheme : string : url scheme, e.g., 'http'

    Returns:
        string : normalized port data.

    """
    if not port.isdigit():
        return port
    port = str(int(port))
    if DEFAULT_PORT.get(scheme) == port:
        return ""
    return port



"""URL scheme normalization."""

from __future__ import annotations

DEFAULT_SCHEME = "https"


def normalize_scheme(scheme: str) -> str:
    """Normalize scheme part of the url.

    Params:
        scheme : string : url scheme, e.g., 'https'

    Returns:
        string : normalized scheme data.

    """
    return scheme.lower()
